#include "global.h"

Epp::Epp(ScreenOutput *disp, AudioOutput *au, Keyboard *kb, Joystick *js, Mouse *mo) {

    log("Creating front end handler.");

	machine=0;
    display=disp;
    audio=au;
    keyboard=kb;
    joystick=js;
    mouse=mo;
}

Epp::~Epp() {

	if(machine) delete machine;
}

void Epp::setMachine(Machine *mach) {

	machine=mach;
}

Machine *Epp::getMachineByName(const char *name, Epp *e, ScreenOutput *disp, AudioOutput *au, Keyboard *kb, Joystick *js, Mouse *mo) {

    if (!strcmp(name, "spectrum")) {

   		return new Spectrum(e, disp, au, kb, js, 0);
    }
   	else if (!strcmp(name, "invaders")) {

   		return new Invaders(e, disp, au, kb, js);
    }
   	else if (!strcmp(name, "lrescue")) {

   		return new LunarRescue(e, disp, au, kb, js);
    }
   	else if (!strcmp(name, "phoenix")) {

   		return new Phoenix(e, disp, au, kb, js);
    }
   	else if (!strcmp(name, "pleiads")) {

   		return new Pleiads(e, disp, au, kb, js);
    }
   	else if (!strcmp(name, "centipede")) {

   		return new Centipede(e, disp, au, kb, js, mo);
    }
    return 0;

}

bool Epp::checkInterface(void) {

    // Check for exiting the emulator
   	if(keyboard->keyPressed(Keyboard::k_esc)) {

   		return true;
   	}

    // F1 will do a reset
   	if(keyboard->keyPressed(Keyboard::k_f1)) machine->reset();

    // F11 shows framerate;
   	if(keyboard->keyClicked(Keyboard::k_f11)) display->toggleFPS();

    return false;
}

struct Settings {

	char *name;
	uint8 current;
    char *option[8];
};

struct MachineInfo {

	char *name;
    char *description;
    Settings setting[10];
};

// I want to put this in the Epp class, but the linker's having none of it.

static MachineInfo info[] =

{
    {
        "centipede",
		"Centipede\n---------\n\nAtari 1980\n\nShoot the bugs and\nmushrooms.  Use either\njoystick or trackball.",
        {
    		{ "Language", 0, { "English", "German", "French", "Spanish", NULL } },
            { "Lives", 1, { "2", "3", "4", "5", NULL } },
            { "Bonus", 1, { "10000", "12000", "15000", "20000", NULL } },
            { "Level", 1, { "Hard", "Easy", NULL } },
            { NULL }
        }
    },

    {
    	"invaders",
        "Space Invaders\n--------------\n\nTaito 1979?\n\nClassic arcade shooter\nfrom Taito.  Shoot\nyour way way though\nwave after wave of\nalien Space Invaders.",
        {
    		{ "Lives", 0, { "3", "4", "5", "6", NULL } },
            { "Bonus", 0, { "1500", "1000", NULL } },
            { NULL }
        }
    },

    {
    	"lrescue",
        "Lunar Rescue\n------------\n\nTaito 1979\n\nRuns on Space Invaders\nhardware.  Rescue the\nstranded astronauts from\nthe moon.",
        {
    		{ "Lives", 0, { "3", "4", "5", "6", NULL } },
            { "Bonus", 0, { "1500", "1000", NULL } },
            { NULL }
        }
    },

    {
    	"phoenix",
        "Phoenix\n-------\n\nTaito 1980\n\nAmong the first wave\nof multi-stage shoot-\nem-ups.  Colourful,\nscrolling graphics and\nlevel transitions.",
        {
    		{ "Lives", 0, { "3", "4", "5", "6", NULL } },
            { "Bonus", 0, { "3000", "4000", "5000", "6000", NULL } },
            { NULL }
        }
    },

    {
    	"pleiads",
        "Pleiads\n-------\n\nCentauri 1981\n\nRuns on the same\nhardware as Phoenix.",
        {
    		{ "Lives", 0, { "3", "4", "5", "6", NULL } },
            { "Bonus", 0, { "3000", "4000", "5000", "6000", NULL } },
            { NULL }
        }
    },

    {
    	"spectrum",
	    "Spectrum\n--------\n\nSinclair Research 1982\n\nPopular home computer of\nthe early 80s.  Created\nby Sinclair Research.\n\nYou can specify a game\nsnapshot on the command\nline.",
        {
    		{ NULL }
        }
    },

    // End of list
    {
    	NULL
   	}
};

void Epp::start(void) {

	if(machine) {

    	// We have a machine, so run it

    	machine->emulate();
        return;
    }
    else {

    	// No machine created on command line, so we start the front end

        setupAudio();
        uint16 count=setupScreen();
        uint16 machno=0;
		bool infoflag=false;
        uint8 infocount=0;
        uint8 options=0;

        //cerr << nohope[0].name << nohope[0].settings[1].name;

        do {

			display->drawText(display->getScreen(), 20, 70+(10*machno), "->", 255);

            if(keyboard->keyPressed(Keyboard::k_up) && machno) {

				display->drawText(display->getScreen(), 20, 70+(10*machno), "  ", 255);
				machno--;
			    audio->playWaveform(samples[0], channel);

                infoflag=false;
                infocount=0;

                while(keyboard->keyPressed(Keyboard::k_up));
            }

            if(keyboard->keyPressed(Keyboard::k_down) && machno<count) {

				display->drawText(display->getScreen(), 20, 70+(10*machno), "  ", 255);
				machno++;
        	    audio->playWaveform(samples[0], channel);

                infoflag=false;
                infocount=0;

                while(keyboard->keyPressed(Keyboard::k_down));
            }

            if(keyboard->keyPressed(Keyboard::k_enter)) {

                freeAudio();

            	machine=getMachineByName(info[machno].name, this, display, audio, keyboard, joystick, mouse);
                machine->emulate();
                delete machine;

                setupAudio();
                setupScreen();

                infoflag=false;
	            infocount=0;
            }

            if(infoflag) {

            	uint8 option=255;

                if(keyboard->keyClicked(Keyboard::k_0) && options >= 1) option=0;
                if(keyboard->keyClicked(Keyboard::k_1) && options >= 2) option=1;
                if(keyboard->keyClicked(Keyboard::k_2) && options >= 3) option=2;
                if(keyboard->keyClicked(Keyboard::k_3) && options >= 4) option=3;
                if(keyboard->keyClicked(Keyboard::k_4) && options >= 5) option=4;
                if(keyboard->keyClicked(Keyboard::k_5) && options >= 6) option=5;
                if(keyboard->keyClicked(Keyboard::k_6) && options >= 7) option=6;
                if(keyboard->keyClicked(Keyboard::k_7) && options >= 8) option=7;
                if(keyboard->keyClicked(Keyboard::k_8) && options >= 9) option=8;
                if(keyboard->keyClicked(Keyboard::k_9) && options >= 10) option=9;

                if(option<255) {

                	if(info[machno].setting[option].option[(info[machno].setting[option].current)+1]) {

						(info[machno].setting[option].current)++;
                    }
                    else {

						(info[machno].setting[option].current)=0;
                    }

                    teleprint(418+(16*8), 280+18+(option*9), "        ");
                    teleprint(418+(16*8), 280+18+(option*9), info[machno].setting[option].option[info[machno].setting[option].current]);
                }
            }

            if(infocount==200 && !infoflag) {

            	// Iccy hard-coded numbers alert!

            	infoflag=true;
                wipe(416, 48, 416+4+(24*8), 48+4+(18*9));
                teleprint(418, 50, info[machno].description);

                wipe(416, 278, 416+4+(24*8), 278+4+(18*9));
                teleprint(418, 280, "Options:");

                int16 y=280+18;
                char num[2];

                for(options=0; info[machno].setting[options].name; options++) {

                    teleprint(418, y, itoa(options, num, 10));
                    teleprint(418+8, y, ":");
                	teleprint(418+24, y, info[machno].setting[options].name);
                    teleprint(418+(16*8), y, info[machno].setting[options].option[info[machno].setting[options].current]);

                    y+=9;
                }
            }
            else {

            	infocount++;
            }

            display->update();
			audio->update();

        } while(!keyboard->keyPressed(Keyboard::k_esc));

        freeAudio();
        return;
    }
}

uint8 Epp::getSetting(char *machname, uint8 setting) {

    uint16 machno;

	for(machno=0; strcmp(machname, info[machno].name); machno++);

    return info[machno].setting[setting].current;
}

void Epp::resetLog(void) {

    // Create the log file

   	ofstream output("Epplog.txt", ios::out|ios::trunc);
    output << "E++ logging started...\n\n";
	output.close();
}

void Epp::log(char *text) {

	// Write to the log file

   	ofstream output("Epplog.txt", ios::out|ios::app);
    output << text << "\n";
	output.close();
}

uint16 Epp::setupScreen(void) {

	display->setMode(640, 480, 200);

    void *bg=display->loadBitmap("bitmaps/menu.pcx");

    display->copyBitmap(bg, display->getScreen(), 0, 0, ScreenOutput::solid);
    display->freeBitmap(bg);

    display->setPalette(255, 0x00FF00);
    display->setPalette(254, 0xFFFFFF);
    display->setPalette(253, 0x008800);

    for(int i=0; i<60*2; i++) {

    	display->update();
    }

    teleprint(20, 50, "Welcome.  Please select machine:");

    uint16 count;

   	for(count=0; info[count].name; count++) {

		teleprint(38, 70+(10*count), info[count].name);
	}

    return count-1;
}

void Epp::setupAudio(void) {

	channel=audio->openChannel(255, 127);
    samples[0]=audio->loadSample("samples/frontend.0");
    samples[1]=audio->loadSample("samples/frontend.1");
}

void Epp::freeAudio(void) {

	audio->closeChannel(channel);
    audio->freeSample(samples[0]);
    audio->freeSample(samples[1]);
}

void Epp::teleprint(int16 x, int16 y, char *text) {

    int16 xpos=x, ypos=y;

    for(int i=0; text[i]; i++) {

        if(text[i] == '\n') {

        	xpos=x;
            ypos+=9;

            display->update();
        }
        else {

			display->drawChar(display->getScreen(), xpos, ypos, text[i], 254);

            if(!audio->channelActive(channel)) {

            	audio->playWaveform(samples[1], channel);
            }

		    audio->update();

        	display->update();

			display->drawChar(display->getScreen(), xpos, ypos, text[i], 255);
            audio->update();
        	display->update();

	        xpos+=8;
        }

    }
}

void Epp::wipe(int16 x1, int16 y1, int16 x2, int16 y2) {

	for(int ypos=y1+6; ypos<=y2; ypos+=4) {

    	display->drawLine(display->getScreen(), x1, ypos-6, x2, ypos-6, 0);
    	display->drawLine(display->getScreen(), x1, ypos-5, x2, ypos-5, 0);
    	display->drawLine(display->getScreen(), x1, ypos-4, x2, ypos-4, 0);
    	display->drawLine(display->getScreen(), x1, ypos-3, x2, ypos-3, 0);
    	display->drawLine(display->getScreen(), x1, ypos-2, x2, ypos-2, 253);
    	display->drawLine(display->getScreen(), x1, ypos-1, x2, ypos-1, 255);
    	display->drawLine(display->getScreen(), x1, ypos-0, x2, ypos-0, 254);

        display->update();
        audio->update();
    }

    display->drawLine(display->getScreen(), x1, y2-2, x2, y2-2, 0);
    display->drawLine(display->getScreen(), x1, y2-1, x2, y2-1, 0);
    display->drawLine(display->getScreen(), x1, y2-0, x2, y2-0, 0);
    display->update();
}
