#include "../../global.h"

SinclairVideo::SinclairVideo(Machine *hw, ScreenOutput *disp, uint8 *cr) : VideoDevice(hw, disp, cr) {

    uint16 mod, y, pos=0;

    flash = false;
    border = 0;
    newborder = false;
    flashflag = false;

    for(int i=6144; i<6912; i++) {

        dirty[i]=true;
    }

    // The order of the lines is really odd, so we pre-calculate
    // them, so that redraws are faster.

    for (mod=0; mod <8; mod++) {
        for(y=0; y<63; y+=8) {
            positions[y+mod]=pos;
            pos+=32;
        }
    }
    for (mod=0; mod <8; mod++) {
        for(y=64; y<127; y+=8) {
            positions[y+mod]=pos;
            pos+=32;
        }
    }
    for (mod=0; mod <8; mod++) {
        for(y=128; y<191; y+=8) {
            positions[y+mod]=pos;
            pos+=32;
        }
    }
}

void SinclairVideo::setupDisplay(void) {

    display->setMode(256+48, 192+48, 50);

    display->setPalette(0, 0x000000);
    display->setPalette(1, 0x0000cc);
    display->setPalette(2, 0xcc0000);
    display->setPalette(3, 0xcc00cc);
    display->setPalette(4, 0x00cc00);
    display->setPalette(5, 0x00cccc);
    display->setPalette(6, 0xcccc00);
    display->setPalette(7, 0xcccccc);
    display->setPalette(8, 0x000000);
    display->setPalette(9, 0x0000FF);
    display->setPalette(10, 0xFF0000);
    display->setPalette(11, 0xFF00FF);
    display->setPalette(12, 0x00FF00);
    display->setPalette(13, 0x00FFFF);
    display->setPalette(14, 0xFFFF00);
    display->setPalette(15, 0xFFFFFF);

    display->clearBitmap(display->getScreen(), border);

    // If a snapshot has been loaded, we can't guarantee that the
    // screen will be redrawn, so fill the video RAM now.

    for(int i=0; i<6912; i++) {

        videoRAM[i]=machine->readMem(this, i+16384);
    }
}

void SinclairVideo::update(void) {

    int x, y;
    int attrpos, pixelpos;
    uint8 data, attr, bg, fg;

    // We have a 24 pixel border

    for (y=0+24; y<192+24; y++) {

        // Addresses of the attribute and pixel data in video RAM

        attrpos = 6144+32*((y-24)>>3);
        pixelpos = positions[y-24];

        for(x=0+24; x<256+24; x+=8) {

            attr = videoRAM[attrpos];

            // We need to redraw if:
            //
            // * The attribute has changed.
            // * The pixel data has changed, or...
            // * The flash attribute is set, and we have just inverted
            //   the flash state.

            if(dirty[attrpos] || dirty[pixelpos] || ((attr&0x80) && flashflag)) {

                data=videoRAM[pixelpos];

                dirty[pixelpos]=0;

                // Get the base background and foreground colours

                if(flash && (attr & 0x80)) {

                    bg=(attr&0x07);
                    fg=((attr&0x38)>>3);
                }
                else {

                    fg=(attr&0x07);
                    bg=((attr&0x38)>>3);
                }

                // If bright, then use the bright versions

                if(attr&0x40) {

                    fg+=8;
                    bg+=8;
                }

                // Plot the pixels

                display->setScreenPixel(x, y, (data&0x80)==0?bg:fg);
                display->setScreenPixel(x+1, y, (data&0x40)==0?bg:fg);
                display->setScreenPixel(x+2, y, (data&0x20)==0?bg:fg);
                display->setScreenPixel(x+3, y, (data&0x10)==0?bg:fg);
                display->setScreenPixel(x+4, y, (data&0x08)==0?bg:fg);
                display->setScreenPixel(x+5, y, (data&0x04)==0?bg:fg);
                display->setScreenPixel(x+6, y, (data&0x02)==0?bg:fg);
                display->setScreenPixel(x+7, y, (data&0x01)==0?bg:fg);
            }
			// Get the next pixel and attribute bytes
            attrpos++;
            pixelpos++;
        }
    }

    flashflag=false;

    for(int i=6144; i<6912; i++) {

        dirty[i]=false;
    }

    // We may need to redraw the border, if it's changed

    if (newborder) {

        newborder=false;

        for (y=0; y<24; y++) {

            for (x=0; x < 256+48; x++) {

                display->setScreenPixel(x, y, border);
            }
        }


        for (y=24; y<192+24; y++) {

            for (x=0; x<24; x++) {

                display->setScreenPixel(x, y, border);
            }
        }

        for (y=0; y<24; y++) {

            for (x=0; x < 256+48; x++) {

                display->setScreenPixel(x, y+24+192, border);
            }
        }


        for (y=24; y<192+24; y++) {

            for (x=0; x<24; x++) {

                display->setScreenPixel(x+24+256, y, border);
            }
        }

    }

    display->update();
}

void SinclairVideo::writeVideoRAM(uint16 addr, uint8 data) {

    // Try to get a bit more speed with both a dirty buffer and
    // a local copy of the video RAM.

    dirty[addr]=1;
    videoRAM[addr]=data;
}

void SinclairVideo::invert(void) {

    flash=!flash;
    flashflag=true;
}
