#include "../global.h"
#include <stdio.h>

ScreenOutput::ScreenOutput() {

    Epp::log("Openning graphic device.");

    fps=0;
    width=height=0;
    xmod=ymod=0;
    framerate=0;
    bitmap=0;
    fpsflag=false;

	// Give ticks an initial value, and start the clock

    ticks=uclock();
}

ScreenOutput::~ScreenOutput() {

    Epp::log("Closing graphic device.");

    if(bitmap) freeBitmap(bitmap);
    set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
}

void ScreenOutput::setMode(const uint16 wid, const uint16 hei, const uint16 fr) {

    Epp::log("Setting screen resolution.");

    width=wid;
    height=hei;
    framerate=fr;

    if(bitmap) freeBitmap(bitmap);

    bitmap=create_bitmap(width, height);
	clear_to_color(bitmap, 0);

    // We use Vesa 2 - is this unreasonable?  Makes drawing faster

    if(wid>320) {

    	if (set_gfx_mode(GFX_VESA2L, 640, 480, 0, 0) &&
    	    set_gfx_mode(GFX_VESA2B, 640, 480, 0, 0)) {

    		Epp::log("Unable to set VESA graphics mode.\nYou MUST have VESA 2.0 support (640x480 resolution)!");
        	exit(1);
    	}
        xmod=ymod=0;
    }
    else {

    	int xres=320, yres=240;

        if(height>240) {

            xres=400;
            yres=300;
        }

    	xmod=(xres-width)/2;
        ymod=(yres-height)/2;

    	if (set_gfx_mode(GFX_VESA2L, xres, yres, 0, 0) &&
    	    set_gfx_mode(GFX_VESA2B, xres, xres, 0, 0)) {

    		Epp::log("Unable to set VESA graphics mode.\nYou MUST have VESA 2.0 support!");
			Epp::log("Trying ModeX mode.  Warning ModeX support is very slow at the moment!\nYou should get a VESA 2.0 driver if possible.");

            if(set_gfx_mode(GFX_MODEX, 360, 270, 0, 0)) {

            	Epp::log("Couldn't open ModeX either!!!");
    			exit(1);
            }
            else {

    			xmod=(360-width)/2;
		        ymod=(270-height)/2;
            }
    	}
    }
}

void ScreenOutput::update(void) {

    uclock_t temp;

    if(fpsflag) drawFPS();
    blit(bitmap, screen, 0, 0, xmod, ymod, width, height);

    do {

        // Get the clock, calculate frames/sec repeat as required
    	temp=uclock();
    	fps=UCLOCKS_PER_SEC/(temp-ticks);
    }
    while(fps > framerate);

    // I can use this to set my monitor positions!
    // clear_to_color(screen, 1);

    ticks=temp;
}

void ScreenOutput::toggleFPS(void) {

	fpsflag=!fpsflag;
}

void ScreenOutput::drawFPS(void) {

   	unsigned char outp[5];
   	sprintf((char *)outp, "%d ", fps);
   	textout(bitmap, font, outp, 0, 0, 1);
}

void ScreenOutput::setPixel(void *bm, const int16 x, const int16 y, const uint8 colour) {

        ((BITMAP *)bm)->line[y][x]=colour;
}

void ScreenOutput::drawChar(void *bm, const int16 x, const int16 y, const char c, const uint8 colour) {

    unsigned char outp[2];
    sprintf((char *)outp, "%c", c);
	textout((BITMAP *)bm, font, outp, x, y, colour);
}

void ScreenOutput::drawText(void *bm, const int16 x, const int16 y, char *text, const uint8 colour) {

	textout((BITMAP *)bm, font, (unsigned char *)text, x, y, colour);
}

void ScreenOutput::drawLine(void *bm, const int16 x, const int16 y, const int16 x2, const int16 y2, const uint8 colour) {

    line((BITMAP *)bm, x, y, x2, y2, colour);
}

void ScreenOutput::setPalette(const uint8 index, const uint32 value) {

	RGB colour;

    // PC values 0..63, but 0..255 seems more reasonable to me, so
    // we divide down in this implementation.
    colour.r=((value >> 16) & 0xFF) >> 2;
    colour.g=((value >> 8) & 0xFF) >> 2;
    colour.b=((value >> 0) & 0xFF) >> 2;

    set_color((int)index, &colour);
}

void ScreenOutput::clearBitmap(void *bm, const uint8 colour) {

	clear_to_color((BITMAP *)bm, colour);
}

uint16 ScreenOutput::getFPS(void) const {

	return fps;
}

void *ScreenOutput::createBitmap(const uint16 wid, const uint16 hei) {

	return create_bitmap(wid, hei);
}

void *ScreenOutput::loadBitmap(char *file) {

	RGB pal[256];
    BITMAP *bm;

    if(!(bm=load_pcx(file, pal))) {

    	Epp::log("Unable to load bitmap file.");
        exit(1);
    }

    set_palette(pal);
    return bm;
}

void ScreenOutput::createSpriteBank(uint8 *data[], void *bank[], const uint16 count, const uint16 wid, const uint16 hei) {

    int pos;

	for(int i=0; i<count; i++) {

    	bank[i]=createBitmap(wid, hei);
        pos=0;

        for(int y=0; y<hei; y++) {

        	for(int x=0; x<wid; x++) {

        		putpixel((BITMAP *)bank[i], x, y, *(data[i]+pos));
                pos++;
            }
        }
    }
}

void ScreenOutput::freeSpriteBank(void *bank[], const uint16 count) {

	for(int i=0; i<count; i++) {

    	freeBitmap(bank[i]);
    }
}

void ScreenOutput::copyBitmap(void *src, void *dest, const int16 x, const int16 y, const uint8 trans) {

 	switch (trans) {

     	case solid:
        blit((BITMAP *)src, (BITMAP *)dest, 0, 0, x, y, ((BITMAP *)src)->w, ((BITMAP *)src)->h);
        break;

     	case transparent:
        draw_sprite((BITMAP *)dest, (BITMAP *)src, x, y);
        break;

        default:
        break;
    }
}

void ScreenOutput::copyWrappedBitmap(void *src, void *dest, const int16 x, const int16 y, int16 ix, int16 iy, const uint8 trans) {

    int swid = ((BITMAP *)src)->w;
    int shei = ((BITMAP *)src)->h;

    BITMAP *temp=create_bitmap(swid, shei);

    if(ix<0) ix+=(swid);
    if(iy<0) iy+=(shei);

    blit((BITMAP *)src, temp, 0, 0, ix, iy, swid, shei);
    if(ix) blit((BITMAP *)src, temp, 0, 0, ix-swid, iy, swid, shei);
    if(iy) blit((BITMAP *)src, temp, 0, 0, ix, iy-shei, swid, shei);
    if(ix&&iy) blit((BITMAP *)src, temp, 0, 0, ix-swid, iy-shei, swid, shei);

	switch (trans) {

     	case solid:
        blit(temp, (BITMAP *)dest, 0, 0, x, y, swid, shei);
        break;

     	case transparent:
        draw_sprite(temp, (BITMAP *)dest, x, y);
        break;

        default:
        break;
    }

    destroy_bitmap(temp);
}

void ScreenOutput::drawSprite(void *src, void *dest, const int16 x, const int16 y, const uint8 trans, const uint8 *colourtable, const bool xflip, const bool yflip) {

    // INCOMPLETE - can't flip solid sprites yet!

	int wid=((BITMAP *)src)->w;
    int hei=((BITMAP *)src)->h;

    BITMAP *temp=create_bitmap(wid, hei);

    if(colourtable) {

    	for(int y2=0; y2<hei; y2++) {

    		for(int x2=0; x2<wid; x2++) {

    			temp->line[y2][x2]=colourtable[((BITMAP *)src)->line[y2][x2]];
    		}
    	}
    }

 	switch (trans) {

     	case solid:
        blit((BITMAP *)src, (BITMAP *)dest, 0, 0, x, y, ((BITMAP *)src)->w, ((BITMAP *)src)->h);
        break;

     	case transparent:
        if(xflip && yflip) {

        	draw_sprite_vh_flip((BITMAP *)dest, temp, x, y);
        }
        else if (xflip) {

        	draw_sprite_h_flip((BITMAP *)dest, temp, x, y);
        }
        else if (yflip) {

        	draw_sprite_v_flip((BITMAP *)dest, temp, x, y);
        }
        else {

        	draw_sprite((BITMAP *)dest, temp, x, y);
        }
        break;

        default:
        break;
    }

    destroy_bitmap(temp);
}

void ScreenOutput::freeBitmap(void *bm) {

	destroy_bitmap((BITMAP *)bm);
}
